import 'dart:async';

import 'package:flutter/services.dart';

/// Wraps MMKV on iOS) and  Android, providing
/// a persistent store for simple data.
class FlutterMmkvCache {
  static const MethodChannel _channel =
  const MethodChannel('flutter_mmkv_cache');

  static const String KEY = 'key';
  static const String VALUE = 'value';

  static const String _prefix = 'flutter_mmkv_cache_';
  static FlutterMmkvCache _instance;
  final Map<String, Object> _mmkvCache;

  FlutterMmkvCache._(this._mmkvCache);

  static Future<FlutterMmkvCache> getInstance() async {
    if (_instance == null) {
      final Map<String, Object> mmkvMap = <String, Object>{};
      _instance = FlutterMmkvCache._(mmkvMap);
    }
    return _instance;
  }

  /// Reads a value from persistent storage, throwing an exception if it's not a bool.
  Future<bool> getBool(String key) async {
    if (_mmkvCache.containsKey(key)) {
      return _mmkvCache[key];
    } else {
      final Map<String, dynamic> params = <String, dynamic>{
        KEY: '$_prefix$key'
      };
      return await _channel
          .invokeMethod('getBool', params)
          .then<bool>((result) {
        _mmkvCache[key] = result;
        return result;
      });
    }
  }

  /// Reads a value from persistent storage, throwing an exception if it's not an int.
  Future<int> getInt(String key) async {
    if (_mmkvCache.containsKey(key)) {
      print('get value from catch');
      return _mmkvCache[key];
    } else {
      print('get value from channel');
      final Map<String, dynamic> params = <String, dynamic>{
        KEY: '$_prefix$key'
      };
      return await _channel.invokeMethod('getInt', params).then<int>((result) {
        _mmkvCache[key] = result;
        return result;
      });
    }
  }

  /// Reads a value from persistent storage, throwing an exception if it's not a long.
  Future<int> getLong(String key) async {
    if (_mmkvCache.containsKey(key)) {
      return _mmkvCache[key];
    } else {
      final Map<String, dynamic> params = <String, dynamic>{
        KEY: '$_prefix$key'
      };
      return await _channel.invokeMethod('getLong', params).then<int>((result) {
        _mmkvCache[key] = result;
        return result;
      });
    }
  }

  /// Reads a value from persistent storage, throwing an exception if it's not a double.
  Future<double> getDouble(String key) async {
    if (_mmkvCache.containsKey(key)) {
      return _mmkvCache[key];
    } else {
      final Map<String, dynamic> params = <String, dynamic>{
        KEY: '$_prefix$key'
      };
      return await _channel
          .invokeMethod('getDouble', params)
          .then<double>((result) {
        _mmkvCache[key] = result;
        return result;
      });
    }
  }

  /// Reads a value from persistent storage, throwing an exception if it's not a string.
  Future<String> getString(String key) async {
    if (_mmkvCache.containsKey(key)) {
      return _mmkvCache[key];
    } else {
      final Map<String, dynamic> params = <String, dynamic>{
        KEY: '$_prefix$key'
      };
      return await _channel
          .invokeMethod('getString', params)
          .then<String>((result) {
        _mmkvCache[key] = result;
        return result;
      });
    }
  }

  /// Saves a boolean [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [removeByKey()] on the [key].
  Future<bool> setBool(String key, bool value) => _setValue('Bool', key, value);

  /// Saves an integer  [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [removeByKey()] on the [key].
  Future<bool> setInt(String key, int value) => _setValue('Int', key, value);

  /// Saves a long integer  [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [removeByKey()] on the [key].
  Future<bool> setLong(String key, int value) => _setValue('Long', key, value);

  /// Saves a double  [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [removeByKey()] on the [key].
  Future<bool> setDouble(String key, double value) =>
      _setValue('Double', key, value);

  /// Saves a string  [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [removeByKey()] on the [key].
  Future<bool> setString(String key, String value) =>
      _setValue('String', key, value);

  /// Removes an entry from persistent storage.
  Future<bool> removeByKey(String key) => _setValue(null, key, null);

  Future<bool> _setValue(String valueType, String key, Object value) async {
    final Map<String, dynamic> params = <String, dynamic>{KEY: '$_prefix$key'};

    if (value == null) {
      _mmkvCache.remove(key);
      return await _channel
          .invokeMethod('removeByKey', params)
          .then<bool>((dynamic result) => result);
    } else {
      _mmkvCache[key] = value;
      params[VALUE] = value;
      return await _channel
          .invokeMethod('set$valueType', params)
          .then<bool>((dynamic result) => result);
    }
  }

  /// Get ValueSize For Key
  Future<int> getValueSizeForKey(String key) async {
    final Map<String, dynamic> params = <String, dynamic>{KEY: '$_prefix$key'};
    return await _channel
        .invokeMethod('getValueSizeForKey', params)
        .then<int>((result) => result);
  }

  /// Get all keys from mmkv
  Future<List<String>> allKeys() async {
    return await _channel
        .invokeMethod('allKeys')
        .then<List<String>>((result) => List<String>.from(result));
  }

  /// Completes with true once the user preferences for the app has been cleared.
  Future<bool> clear() async {
    _mmkvCache.clear();
    return await _channel.invokeMethod('clear');
  }
}

/// MMKV Cache with expire support
class ExpiredCache {
  static _getPrefix(String key) => 'expired_$key';
  static FlutterMmkvCache _mmkv;
  static getInstance() async {
    if (ExpiredCache._mmkv != null) {
      return ExpiredCache._mmkv;
    }
    ExpiredCache._mmkv = await FlutterMmkvCache.getInstance();
    return ExpiredCache._mmkv;
  }

  /// add item to cache; ```expired: hours``` defaults to 0
  static Future<bool> add(String key, String value, {int expired = 0}) async {
    final mmkv = await ExpiredCache.getInstance();
    final timespan =
        DateTime.now().add(new Duration(hours: expired)).millisecondsSinceEpoch;

    /// add timespan to value so we can use it later.
    final v = '$timespan:$value';
    return mmkv.setString(_getPrefix(key), v);
  }

  /// get item from cache if its not exists or has expired return null
  static Future<String> get(String key) async {
    final mmkv = await ExpiredCache.getInstance();
    final value = await mmkv.getString(_getPrefix(key));
    if (value == null || value.isEmpty) {
      return '';
    }
    final index = value.indexOf(':');
    if (index < 0) {
      return '';
    }

    /// get timespan
    final timespan = value.substring(0, index);
    final diff =
    int.parse(timespan).compareTo(DateTime.now().millisecondsSinceEpoch);
    if (diff < 1) {
      await mmkv.removeByKey(_getPrefix(key));
      return '';
    }

    /// get user data except `:`
    final v = value.substring(index + 1);
    return v;
  }

  /// clear item by key or clear all if key not provided
  static Future<bool> clear({String key = ""}) async {
    final mmkv = await ExpiredCache.getInstance();
    if (key.isNotEmpty) {
      return mmkv.removeByKey(key);
    }
    return mmkv.clear();
  }
}
